#include "eviveScratchFirmware.h"

Servo eviveServoMotors[8];                          //Servo Motor
Stepper eviveStepperMotor(200, 28, 29, 30, 31);     //Stepper Motor
DHT eviveDHTHumidityTemperature;                                //DHT Sensor
bool eviveDHTFlag[28];
Adafruit_MPR121 capacitiveTouch = Adafruit_MPR121();
bool eviveTouchFlag = true;
uint16_t eviveTouchLastTouched = 0;
uint16_t eviveTouchCurrentTouched = 0;
Rtc_Pcf8563 eviveRTC;                          //RTC
Motor eviveMotorChannel1, eviveMotorChannel2;                             //DC Motor
int rotataryEnchoder[28][4];                        //Encoder
TFT_ST7735 eviveLCDDisplay = TFT_ST7735();            //TFT

MFRC522 rfid(10, 9); 
byte nuidPICC[4]; 
byte masterKeynuidPICC[4]; 
int masterKeyDefined;       //RFID

char keyvalues[16] = {'0', '1', '2', '3', '4', '5', '6', '7', '8', '9', '*', '#', 'A', 'B', 'C', 'D'};     //Keypad 4x4
char keymap[4][4] = {{'1', '2', '3', 'A'},
  {'4', '5', '6', 'B'},
  {'7', '8', '9', 'C'},
  {'*', '0', '#', 'D'}
};
byte rowPins[4] = {6, 7, 20, 21};
byte colPins[4] = {5, 4, 3, 2};
Keypad myKeypad = Keypad(makeKeymap(keymap), rowPins, colPins, 4, 4);

union {
  byte byteVal[4];
  float floatVal;
  long longVal;
} val;

union {
  byte byteVal[8];
  double doubleVal;
} valDouble;

union {
  byte byteVal[2];
  short shortVal;
} valShort;

int eviveAnalogPins[16] = {A0, A1, A2, A3, A4, A5, A6, A7, A8, A9, A10, A11, A12, A13, A14, A15};
boolean isAvailable = false;
boolean isBluetooth = false;

int len = 52;
char buffer[52];
char bufferBt[52];
byte index = 0;
byte dataLen;
byte modulesLen = 0;
boolean isStart = false;
unsigned char irRead;
char serialRead;

float angleServo = 90.0;
int servo_pins[8] = {0, 0, 0, 0, 0, 0, 0, 0};
unsigned char prevc = 0;
double lastTime = 0.0;
double currentTime = 0.0;
uint8_t keyPressed = 0;
uint8_t command_index = 0;
uint8_t serialFlag = 0;

bool set_rfid()
{
  if (rfid.PICC_IsNewCardPresent() && rfid.PICC_ReadCardSerial() && masterKeyDefined == 0)
  {
    for (byte i = 0; i < 4; i++) masterKeynuidPICC[i] = rfid.uid.uidByte[i];
    masterKeyDefined = 1;
    return 1;
  }
  else return 0;
}

bool get_rfid()
{
  if (rfid.PICC_IsNewCardPresent() && rfid.PICC_ReadCardSerial() && masterKeyDefined != 0) {
    for (byte i = 0; i < 4; i++) nuidPICC[i] = rfid.uid.uidByte[i];
    if (masterKeynuidPICC[0] == nuidPICC[0] ||
        masterKeynuidPICC[1] == nuidPICC[1] ||
        masterKeynuidPICC[2] == nuidPICC[2] ||
        masterKeynuidPICC[3] == nuidPICC[3] )
      return 1;
    else return 0;
  }
  else return 0;
}

void eviveScratchFirmwareSetup() {
  for (int i = 0; i < 28; i++)                              //Humidity and Temperature sensor
    eviveDHTFlag[i] = true;
  eviveRTC.initClock();                                          //RTC
  eviveMotorChannel1.attachMotor(28, 29, 44);                               //Motor 1
  eviveMotorChannel2.attachMotor(30, 31, 45);                               //Motor 2
  for (int i = 0; i < 28; i++)                              //Encoder
    rotataryEnchoder[i][0] = -1;
  SPI.begin();                                              //SPI
  rfid.PCD_Init();                                          //RFID
  Serial3.begin(115200);                                    //Serial
  Serial.begin(115200);
  eviveLCDDisplay.init(INITR_BLACKTAB);                                 //TFT
  eviveLCDDisplay.setRotation(1);
  eviveLCDDisplay.fillScreen(0);
  eviveLCDDisplay.drawBitmap(45, 29, evive_logo, 71, 71, ST7735_BLUE);
}

void eviveScratchFirmwareLoop() {

  //Rotary Encoder
  for (int i = 0; i < 28; i++)
  {
    if (rotataryEnchoder[i][0] != -1)
    {
      int a = digitalRead(i);
      if (a != rotataryEnchoder[i][3])
      {
        if (digitalRead(rotataryEnchoder[i][0]) == a)
        {
          rotataryEnchoder[i][1]++;
          rotataryEnchoder[i][2] = 1;
        }
        else
        {
          rotataryEnchoder[i][1]--;
          rotataryEnchoder[i][2] = 2;
        }
      }
      rotataryEnchoder[i][3] = a;
    }
  }

  currentTime = millis() / 1000.0 - lastTime;

  readSerial();
  if (isAvailable) {
    unsigned char c = serialRead & 0xff;
    if (c == 0x55 && isStart == false) {
      if (prevc == 0xff) {
        index = 1;
        isStart = true;
      }
    } else {
      prevc = c;
      if (isStart) {
        if (index == 2) {
          dataLen = c;
        } else if (index > 2) {
          dataLen--;
        }
        writeBuffer(index, c);
      }
    }
    index++;
    if (index > 51) {
      index = 0;
      isStart = false;
    }
    if (isStart && dataLen == 0 && index > 3) {
      isStart = false;
      parseData();
      index = 0;
    }
  }
}

unsigned char readBuffer(int index) {
  return isBluetooth ? bufferBt[index] : buffer[index];
}

void writeBuffer(int index, unsigned char c) {
  if (isBluetooth) {
    bufferBt[index] = c;
  } else {
    buffer[index] = c;
  }
}

void writeHead() {
  writeSerial(0xff);
  writeSerial(0x55);
}

void writeEnd() {
  if(serialFlag == 0){
    Serial.println();
  }
  if(serialFlag == 3){
    Serial3.println();
  }
#if defined(__AVR_ATmega32U4__)
  Serial1.println();
#endif
}

void writeSerial(unsigned char c) {
  Serial.write(c); 
  Serial3.write(c);
#if defined(__AVR_ATmega32U4__)
  Serial1.write(c);
#endif
}

void readSerial() {
  isAvailable = false;
  if (Serial3.available() > 0) {
    isAvailable = true;
    isBluetooth = false;
    serialRead = Serial3.read();
    serialFlag = 3;
  }
  if (Serial.available() > 0) {
    isAvailable = true;
    isBluetooth = false;
    serialRead = Serial.read();
    serialFlag = 0;
  }
}

void parseData() {
  isStart = false;
  int idx = readBuffer(3);
  command_index = (uint8_t)idx;
  int action = readBuffer(4);
  int device = readBuffer(5);
  switch (action) {
    case GET: {
        writeHead();
        writeSerial(idx);
        readSensor(device);
        writeEnd();
      }
      break;
    case RUN: {
        runModule(device);
        callOK();
      }
      break;

    case START: {
        //start
        callOK();
      }
      break;
  }
  serialFlag =0;
}

void callOK() {
  writeSerial(0xff);
  writeSerial(0x55);
  writeEnd();
}

void sendByte(char c) {
  writeSerial(1);
  writeSerial(c);
}

void sendString(String s) {
  int l = s.length();
  writeSerial(4);
  writeSerial(l);
  for (int i = 0; i < l; i++) {
    writeSerial(s.charAt(i));
  }
}

void sendFloat(float value) {
  writeSerial(0x2);
  val.floatVal = value;
  writeSerial(val.byteVal[0]);
  writeSerial(val.byteVal[1]);
  writeSerial(val.byteVal[2]);
  writeSerial(val.byteVal[3]);
}

void sendShort(double value) {
  writeSerial(3);
  valShort.shortVal = value;
  writeSerial(valShort.byteVal[0]);
  writeSerial(valShort.byteVal[1]);
}

void sendDouble(double value) {
  writeSerial(2);
  valDouble.doubleVal = value;
  writeSerial(valDouble.byteVal[0]);
  writeSerial(valDouble.byteVal[1]);
  writeSerial(valDouble.byteVal[2]);
  writeSerial(valDouble.byteVal[3]);
}

short readShort(int idx) {
  valShort.byteVal[0] = readBuffer(idx);
  valShort.byteVal[1] = readBuffer(idx + 1);
  return valShort.shortVal;
}

float readFloat(int idx) {
  val.byteVal[0] = readBuffer(idx);
  val.byteVal[1] = readBuffer(idx + 1);
  val.byteVal[2] = readBuffer(idx + 2);
  val.byteVal[3] = readBuffer(idx + 3);
  return val.floatVal;
}

long readLong(int idx) {
  val.byteVal[0] = readBuffer(idx);
  val.byteVal[1] = readBuffer(idx + 1);
  val.byteVal[2] = readBuffer(idx + 2);
  val.byteVal[3] = readBuffer(idx + 3);
  return val.longVal;
}

void runModule(int device) {
  int port = readBuffer(6);
  int pin = port;
  switch (device) {
    
    case SETRFID: {
        masterKeyDefined = 0;
        while (!set_rfid()) {}
        //tone(46, 196, 125);
        digitalWrite(29, HIGH);
        delay(500);
        digitalWrite(29, LOW);
        delay(500);
        digitalWrite(29, HIGH);
        delay(500);
        digitalWrite(29, LOW);
      }
      break;
    
    case RUNMOTORA:
      {
        int dir = readBuffer(7);
        int vel = readBuffer(8);
        if (dir == 1 || dir == 2)
          eviveMotorChannel1.moveMotorS(dir, vel);
        else if (dir == 3)
          eviveMotorChannel1.lockMotor();
        else if (dir == 4)
          eviveMotorChannel1.freeMotor();
      }
      break;
    
    case RUNMOTORB:
      {
        int dir = readBuffer(7);
        int vel = readBuffer(8);
        if (dir == 1 || dir == 2)
          eviveMotorChannel2.moveMotorS(dir, vel);
        else if (dir == 3)
          eviveMotorChannel2.lockMotor();
        else if (dir == 4)
          eviveMotorChannel2.freeMotor();
      }
      break;
    
    case STEPPER:
      {
        int steps = readBuffer(7);
        int vel = readBuffer(8);
        int count = readBuffer(9);
        eviveStepperMotor.setSpeed(vel);
        eviveStepperMotor.step(steps);
      }
      break;
    
    case TONE: {
        pinMode(pin, OUTPUT);
        int hz = readShort(7);
        int ms = readShort(9);
        if (ms > 0) {
          tone(pin, hz, ms);
        } else {
          noTone(pin);
        }
      }
      break;
    
    case SERVO: {
      }
      break;
    
    case TIME:
      {
        int hour = pin;
        int minute = readBuffer(7);
        int second = readBuffer(8);
        eviveRTC.setTime(hour, minute, second);
        digitalWrite(29, HIGH);
        delay(500);
        digitalWrite(29, LOW);
      }
      break;
    
    case DATE:
      {
        int date = pin;
        int month = readBuffer(7);
        int year = readBuffer(8);
        int weekday = readBuffer(9);
        eviveRTC.setDate(date, month, year, 0, weekday);
        digitalWrite(31, HIGH);
        delay(500);
        digitalWrite(31, LOW);
      }
      break;

    case DIGITAL: {
        pinMode(pin, OUTPUT);
        int v = readBuffer(7);
        digitalWrite(pin, v);
      }
      break;
    case PWM: {
        pinMode(pin, OUTPUT);
        int v = readBuffer(7);
        analogWrite(pin, v);
      }
      break;

    case SERVO_PIN: {
        int v = readBuffer(7);
        Servo sv = eviveServoMotors[searchServoPin(pin)];
        if (v >= 0 && v <= 180)
        {
          if (!sv.attached())
          {
            sv.attach(pin);
          }
          sv.write(v);
        }
      }
      break;

    case TIMER: {
        lastTime = millis() / 1000.0;
      }
      break;

    case TFTWRITE: {
        String Str_data;
        int len = readBuffer(2) - 3;
        for (int i = 0; i < len; i++) {
          Str_data += (char)readBuffer(6 + i);
        }
        eviveLCDDisplay.print(Str_data);
        Str_data = "";
      }
      break;

    case FILLSCREEN: {
        eviveLCDDisplay.fillScreen(readShort(6));
      }
      break;

    case SETCURSOR: {
        eviveLCDDisplay.setCursor(readBuffer(6), readBuffer(7));
      }
      break;

    case SETTEXTCOLOR: {
        eviveLCDDisplay.setTextColor(readShort(7), readShort(9));
      }
      break;

    case SETTEXTSIZE: {
        eviveLCDDisplay.setTextSize(readBuffer(6));
      }
      break;

    case DRAWLINE: {
        eviveLCDDisplay.drawLine(readBuffer(6), readBuffer(7), readBuffer(8), readBuffer(9), readShort(10));
      }
      break;

    case DRAWVLINE: {
        eviveLCDDisplay.drawFastVLine(readBuffer(6), readBuffer(7), readBuffer(8), readShort(9));
      }
      break;

    case DRAWHLINE: {
        eviveLCDDisplay.drawFastHLine(readBuffer(6), readBuffer(7), readBuffer(8), readShort(9));
      }
      break;

    case FILLRECTANGLE: {
        eviveLCDDisplay.fillRect(readBuffer(6), readBuffer(7), readBuffer(8), readBuffer(9), readShort(10));
      }
      break;

    case FILLROUNDRECTANGLE: {
        eviveLCDDisplay.fillRoundRect(readBuffer(6), readBuffer(7), readBuffer(8), readBuffer(9), readBuffer(10), readShort(11));
      }
      break;

    case FILLCIRCLE: {
        eviveLCDDisplay.fillCircle(readBuffer(6), readBuffer(7), readBuffer(8), readShort(9));
      }
      break;

    case FILLELLIPSE: {
        eviveLCDDisplay.fillEllipse(readBuffer(6), readBuffer(7), readBuffer(8), readBuffer(9), readShort(10));
      }
      break;

    case FILLTRIANGLE: {
        eviveLCDDisplay.fillTriangle(readBuffer(6), readBuffer(7), readBuffer(8), readBuffer(9), readBuffer(10), readBuffer(11), readShort(12));
      }
      break;

    case DRAWRECTANGLE: {
        eviveLCDDisplay.drawRect(readBuffer(6), readBuffer(7), readBuffer(8), readBuffer(9), readShort(10));
      }
      break;

    case DRAWROUNDRECTANGLE: {
        eviveLCDDisplay.drawRoundRect(readBuffer(6), readBuffer(7), readBuffer(8), readBuffer(9), readBuffer(10), readShort(11));
      }
      break;

    case DRAWCIRCLE: {
        eviveLCDDisplay.drawCircle(readBuffer(6), readBuffer(7), readBuffer(8), readShort(9));
      }
      break;

    case DRAWELLIPSE: {
        eviveLCDDisplay.drawEllipse(readBuffer(6), readBuffer(7), readBuffer(8), readBuffer(9), readShort(10));
      }
      break;

    case DRAWTRIANGLE: {
        eviveLCDDisplay.drawTriangle(readBuffer(6), readBuffer(7), readBuffer(8), readBuffer(9), readBuffer(10), readBuffer(11), readShort(12));
      }
      break;
			
		case INITTFT: {
				if(readBuffer(6) == 0){
					eviveLCDDisplay.init(INITR_BLACKTAB); 
				}
				else{
					eviveLCDDisplay.init(INITR_GREENTAB);
				}
				eviveLCDDisplay.setRotation(1);
		}
		break;

  }
}

int searchServoPin(int pin) {
  for (int i = 0; i < 8; i++) {
    if (servo_pins[i] == pin) {
      return i;
    }
    if (servo_pins[i] == 0) {
      servo_pins[i] = pin;
      return i;
    }
  }
  return 0;
}

void readSensor(int device) {
  float value = 0.0;
  int port, slot, pin;
  port = readBuffer(6);
  pin = port;
  switch (device) {

    case  DIGITAL: {
        pinMode(pin, INPUT);
        sendFloat(digitalRead(pin));
      }
      break;
    
    case  IROBSTACLE: {
        pinMode(pin, INPUT);
        sendFloat(!digitalRead(pin));
      }
      break;
    
    case  ANALOG: {
        pin = eviveAnalogPins[pin];
        pinMode(pin, INPUT);
        sendFloat(analogRead(pin));
      }
      break;
    
    case  PULSEIN: {
        int pw = readShort(7);
        pinMode(pin, INPUT);
        sendShort(pulseIn(pin, HIGH, pw));
      }
      break;
    
    case ULTRASONIC_ARDUINO: {
        int trig = readBuffer(6);
        int echo = readBuffer(7);
        pinMode(trig, OUTPUT);
        digitalWrite(trig, LOW);
        delayMicroseconds(2);
        digitalWrite(trig, HIGH);
        delayMicroseconds(10);
        digitalWrite(trig, LOW);
        pinMode(echo, INPUT);
        sendFloat(pulseIn(echo, HIGH, 30000) / 58.0);
      }
      break;
    
    case TIMER: {
        sendFloat((float)currentTime);
      }
      break;

    case DHTH:
      {
        if (eviveDHTFlag[pin] == true)
        {
          eviveDHTHumidityTemperature.setup(pin);
          eviveDHTFlag[pin] = false;
        }
        sendFloat(eviveDHTHumidityTemperature.getHumidity());
      }
      break;
    
    case DHTT:
      {
        if (eviveDHTFlag[pin] == true)
        {
          eviveDHTHumidityTemperature.setup(pin);
          eviveDHTFlag[pin] = false;
        }
        sendFloat(eviveDHTHumidityTemperature.getTemperature());
      }
      break;
    
    case TOUCH:
      {
        if (eviveTouchFlag == true)
        {
          capacitiveTouch.begin(0x5A);
          eviveTouchFlag = false;
        }
        eviveTouchCurrentTouched = capacitiveTouch.touched();
        sendFloat((eviveTouchCurrentTouched & _BV(pin - 1)) && !(eviveTouchLastTouched & _BV(pin - 1)));
      }
      break;
    
    case NAVKEY:
      {
        int val = analogRead(A11);
        sendFloat( (val > ((pin - 1) * 200 + 175)) && (val < ((pin - 1) * 200 + 275)) );
      }
      break;
    
    case TIME:
      {
        if (pin == 11)
          sendFloat(eviveRTC.getHour());
        else if (pin == 12)
          sendFloat(eviveRTC.getMinute());
        else if (pin == 13)
          sendFloat(eviveRTC.getSecond());
      }
      break;
    
    case DATE:
      {
        if (pin == 21)
          sendFloat(eviveRTC.getDay());
        else if (pin == 22)
          sendFloat(eviveRTC.getMonth());
        else if (pin == 23)
          sendFloat(eviveRTC.getYear());
        else if (pin == 24)
          sendFloat(eviveRTC.getWeekday());
      }
      break;
    
    case ROTENC:
      {
        int pin1 = readBuffer(6);
        int pin2 = readBuffer(7);
        rotataryEnchoder[pin1][0] = pin2;
        rotataryEnchoder[pin1][1] = 0;
        rotataryEnchoder[pin1][2] = 1;
      }
      break;
    
    case ENCPOS:
      {
        sendFloat(rotataryEnchoder[pin][1]);
      }
      break;
    
    case ENCDIR:
      {
        if (rotataryEnchoder[pin][2] == readBuffer(7))
          sendFloat(1);
        else
          sendFloat(0);
      }
      break;
    
    case GETRFID:
      {
        sendFloat(get_rfid());
      }
      break;
    
    case KEYPAD:
      {
        sendFloat((char)myKeypad.getKey() == keyvalues[pin]);
      }
      break;
  }
}
